{-# OPTIONS_GHC -XTypeSynonymInstances  #-}

module Data.GraphRewriting.ExtendedEarl 
where

import qualified Data.Graph.Inductive as Graph (Gr,  empty,context, insNode, insEdge, delEdge, lab, lsuc,suc, pre, nodes,labNodes, labEdges, newNodes, inn, out, gelem, ufold, mkGraph, delNode, delLEdge)
import qualified Data.Graph.Inductive.Query.DFS as GraphQu (reachable)

import Data.List

import qualified Data.IntSet as IntSet (singleton, insert, member)

-- | Representing a variable
type Variable = String

-- | Representing a function symbol
type Symbol = String

-- | Representing a Label
data Label = GVar Variable
           | GFun Symbol 
             deriving (Eq,Show)

-- | Representing a node number
type Node = Int

-- | Representing the position in the successor range
type EdgeLabel = Int

-- | Wanted to use "Graph" here, but this name is already taken, so Earl stands for Graph.                      
--   If you do understand this, your German is pretty outstanding, as "Graph" is an old-fashioned German word for "Earl"
--   or "Count" - but using Count I considered really confusing.
type Earl = Graph.Gr Label EdgeLabel

-- | Representing a context with Ancestors, Node, Label of Node, Successors.
type Context = ([(EdgeLabel,Node)],Node, Label,[(EdgeLabel,Node)])

instance ExtendedEarl Earl where
  toEarl      = id
  insertNode  = einsertNode
  deleteNode  = edeleteNode
  insertEdge  = einsertEdge
  deleteEdge  = edeleteEdge
  deleteLEdge = edeleteLEdge


-- | suitable for a structure which bases on an Earl but only spices it up with additional information.
class ExtendedEarl gr where
  -- | toEarl GR: discards any additional information
  toEarl :: gr -> Earl   
  -- | insertNode GR ln: inserts labeled node ln into underlying Earl GR, whereas implementation can make use of einsertNode.
  insertNode :: gr -> (Node,Label) -> gr
  -- | deleteNode GR n: deletes node n from underlying Earl GR, whereas implementation can make use of edeleteNode.
  deleteNode :: gr -> Node -> gr
  -- | insertEdge GR le: inserts labeled edge le into underlying Earl GR, whereas implementation can make use of einsertEdge.
  insertEdge :: gr -> (Node,Node,EdgeLabel) -> gr
  -- | deleteEdge GR e: deletes ALL edges e from n to n',regardelss of the Edgelabel, from underlying Earl GR, whereas implementation can make use of edeleteEdge.
  deleteEdge :: gr -> (Node, Node) -> gr
  -- | deleteLEdge GR e: deletes the labeled edge from the graph, whereas implementation can make use of edeleteLEdge
  deleteLEdge :: gr ->  (Node, Node, EdgeLabel) -> gr


-------------------------------------------------------------------------------------------------------------------
-- * traversal
-------------------------------------------------------------------------------------------------------------------

-- | foldd f gr n c = folds a given function f over the graph gr starting from n
-- with depht first and reaches every succ just once by maintaining a list of
-- already visited nodes.
foldd :: ExtendedEarl gr => (Node -> c -> c) -> gr -> Node -> c -> c  
foldd f gr n startval = fst (gfold' n (f n startval) (succs gr n) [n]) 
  where gfold' n' c sucs visited = case sucs of -- gfold' :: Node -> c -> [Node] -> [Node] -> (c,[Node])
          []        -> (c,visited)
          si:succis -> gfold' n' c' succis vis'  -- breadth
            where (c', vis')| si `elem` visited = (c,visited)  -- depht
                            | otherwise         = gfold' si (f si c) (succs gr si) (si:visited)


-- | foldc f gr con c = folds a given function f over the graph gr starting from
-- context con with depht first and reaches every succ just once by maintaining
-- a list of already visited contexts.
foldc :: ExtendedEarl gr => (Context -> c -> c) -> gr -> Context -> c -> c
foldc f gr con@(_, n, _,_) startval = fst (foldc' con (f con startval) (IntSet.singleton n))
  where foldc' (_, _, _, []) c vis  = (c, vis) 
        foldc' (ancni, ni, lni, ((_,si):sucni)) c vis = foldc' (ancni, ni, lni, sucni) c' vis'  
          where (c', vis') | si `IntSet.member` vis = (c,vis) 
                           | otherwise     = foldc' (consi) (f consi c) (IntSet.insert si vis)
                                             where consi = (context gr si)
 


-------------------------------------------------------------------------------------------------------------------
-- * misecencellaous or however but still important
-------------------------------------------------------------------------------------------------------------------

-- | freshNode gr: a in gr unused node number
freshNode :: ExtendedEarl gr => gr -> Node
freshNode gr = head $ freshNodes gr 1

-- | freshNodes amount gr: some in gr not yet used node numbers, amount indicates how many of these fresh node numbers are actually wanted   
freshNodes :: ExtendedEarl gr => gr -> Int -> [Node]
freshNodes gr i = Graph.newNodes i (toEarl gr)
                                          
-- | redirect gr n n': redirects precedessors of n to n'. 
--   careful: does not delete n, but n does not have any incoming edges any more.
redirect :: ExtendedEarl gr => gr -> Node -> Node -> gr 
redirect gr n n' | n == n'   = gr
                 | otherwise = deleteLEdges (insertEdges gr edges2n') edges2n
                   where edges2n   = Graph.inn (toEarl gr) n -- all incoming edges of node n in GR
                         edges2n'  = [(anc, n', labl) | (anc, _ , labl) <- edges2n]

-- | copy gr n: copies the given node n in gr i.e. inserts a new node n' such that homomorph gr n n' == true,
--   returns the modified graph and the node number of the copied node n'.                         
--   Careful: The new node n' does not yet have an ancestor!                          
copy :: ExtendedEarl gr => gr -> Node -> (gr, Node)
copy gr n = (insertEdges (insertNode gr (n',label gr n)) edges', n')
            where n'     = freshNode gr
                  edges' = [(n', si, i) | (_,si,i) <- out gr n]

-----------------------------------------------------------------------------------------
-- * inspection
-----------------------------------------------------------------------------------------
-- | context gr n: returns the context of a graph
context :: ExtendedEarl gr => gr -> Node -> Context 
context gr n = Graph.context (toEarl gr) n

-- | nodes gr: all node numbers in the graph gr
nodes :: ExtendedEarl gr => gr -> [Node]
nodes gr = Graph.nodes (toEarl gr)

-- | lnodes gr: all labeled nodes in the graph gr
lnodes :: ExtendedEarl gr => gr -> [(Node,Label)]
lnodes gr = Graph.labNodes (toEarl gr)

-- | leafnodes gr: all leaf nodes of gr, i.e., nodes that do not have any successors
leafnodes :: ExtendedEarl gr => gr -> [Node]
leafnodes gr = [n | n <- nodes gr, Graph.suc (toEarl gr) n == []]

-- | ledges gr: all labeled edges in the graph gr
ledges :: ExtendedEarl gr => gr -> [(Node,Node,EdgeLabel)]
ledges gr = Graph.labEdges (toEarl gr)
            
-- | label gr n: label connected to node n in the graph gr
--   this causes an error if n is not element of the graph.
label :: ExtendedEarl gr =>  gr ->  Node -> Label
label gr n = case Graph.lab (toEarl gr) n of
                   Just l -> l
                   _      -> error ("ExtendedEarl.label: Given Node not part of given Graph (Earl)" ++ show n ++ (show $ toEarl gr))

-- | succs gr n: successor nodes ordered by their edge labels i.e. 1st successor is on 1st position in the list.
succs :: ExtendedEarl gr => gr -> Node -> [Node]
succs gr n = map fst (sortBy edgeLab (Graph.lsuc (toEarl gr) n)) 
  where edgeLab (_, el1) (_,el2) = compare el1 el2   

-- | ithsucc gr n i: returns the i-th successor of Node n in Earl gr, error if there is no ith successor
ithsucc :: ExtendedEarl gr => gr -> Node -> EdgeLabel -> Node
ithsucc gr n i = case find (\(_,_,j) -> i == j) (out gr n) of 
                      Just (_,suci,_) -> suci
                      Nothing         -> error "No ith successor found."
  
-- | out gr n: all outgoing labeled edges of node n in gr
out :: ExtendedEarl gr => gr -> Node -> [(Node,Node, EdgeLabel)]
out gr n = Graph.out (toEarl gr) n 

-- | inn gr n: all incoming labeled edges of Node n in gr 
inn :: ExtendedEarl gr => gr -> Node -> [(Node,Node,EdgeLabel)]
inn gr n = Graph.inn (toEarl gr) n

-- | ancs gr n : direct ancestors of n, unordered 
ancs :: ExtendedEarl gr => gr -> Node -> [Node]
ancs gr n = Graph.pre (toEarl gr) n

-- | size GR: amount of nodes in GR. Please note here: There is no correspondence to the node numbers. So even if the size == 3 this does not
--   indicate that node number 4 is unused. Use freshNode\/freshNodes to get unused node number/s.
size :: ExtendedEarl gr => gr -> Int
size gr = Graph.ufold (\ _ i -> i+1) 0 (toEarl gr) 

-- | isVarNode (G,n): true, if node n represents a variable 
isVarNode :: ExtendedEarl gr => gr -> Node -> Bool
isVarNode gr n = case label gr n of
                      GVar _ -> True
                      _      -> False

-- | isDirectlyShared gr n: indicates whether a node n in gr has more than one ancestor node.
--   Careful: False does not indicate that n might not be shared higher up, 
--   but n does not have more than one directly incoming edge                      
isDirectlyShared :: ExtendedEarl gr => gr -> Node -> Bool
isDirectlyShared gr n = length (inn gr n) > 1

-- | sameLabel (g,n) (g',n'): true, if n in g  and n' in g' have the same label,  
--   e.g. (Fun f) \/= (Fun g),  (Fun f) == (Fun f), (Var _) \/= (Fun _)                   
sameLabel :: ExtendedEarl gr => ExtendedEarl gr2 => (gr,Node) -> (gr2,Node) -> Bool
sameLabel (gr1,n1) (gr2,n2) = label gr1 n1 == label gr2 n2 

-- | homomorph gr n n': returns True if the two nodes n n' in gr
--   have the same label and the same successors. 
--   Attention: Every Node is homomorph to itself i.e. homomorph gr n n == True
--   homomorph (TermGraph tg) n n' == False does not imply term(tg |^ n) /= term (tg |^ n')
homomorph :: ExtendedEarl gr => gr -> Node -> Node -> Bool
homomorph gr n n' = l == l' && ((s \\ s') == [])
   where ((_, _, l, s), (_, _, l', s')) = (Graph.context gr' n, Graph.context gr' n') 
         gr' = toEarl gr

-- | isreachable gr n n': true, if (n==n') or n' is a, not necessarily direct, successor of n. 
isreachable :: ExtendedEarl gr => gr -> Node -> Node -> Bool
isreachable gr n n' = foldd reachedN'  gr n False
     where reachedN' s val = val || (s==n')

-- | reachabl gr n: returns all nodes reachable from n in gr
reachabl :: ExtendedEarl gr => gr -> Node -> [Node]
reachabl gr n = GraphQu.reachable n (toEarl gr) 

-----------------------------------------------------------------------------------------
-- * construction, destruction
-----------------------------------------------------------------------------------------

-- | insertNodes gr lns: insert labeled nodes lns into the graph gr                                                      
insertNodes :: ExtendedEarl gr => gr -> [(Node,Label)] -> gr          
insertNodes = foldl insertNode 

-- | deleteNodes gr ns: deletes nodes ns from the graph gr                                                      
deleteNodes :: ExtendedEarl gr => gr -> [Node] -> gr          
deleteNodes = foldl deleteNode 

-- | insertEdges gr les: inserts labeled edges les into the graph gr                                                      
insertEdges :: ExtendedEarl gr => gr -> [(Node,Node,EdgeLabel)] -> gr          
insertEdges = foldl insertEdge
  
-- | deleteEdges gr edgs: deletes edges egds from the graph gr                                                      
deleteEdges :: ExtendedEarl gr => gr -> [(Node,Node)] -> gr          
deleteEdges = foldl deleteEdge

-- | deleteLEdges gr ledgs: deletes labeled edges ledgs from the graph gr
deleteLEdges :: ExtendedEarl gr => gr -> [(Node,Node,EdgeLabel)] -> gr
deleteLEdges = foldl deleteLEdge 

--------------------------------------------------------------------------
-- * earl based
--------------------------------------------------------------------------

-- | empty: returns an empty Earl
empty :: Earl 
empty = Graph.empty

-- | einsertNode gr ln: inserts the labeled node ln into the Earl gr
einsertNode :: Earl -> (Node,Label) -> Earl
einsertNode gr ln = Graph.insNode ln gr

-- | edeleteNode gr n: deletes node n from the Earl gr
edeleteNode :: Earl -> Node ->Earl
edeleteNode gr n = Graph.delNode n gr

-- | einsertEdge gr le: inserts labeled edge le into graph gr, but doesn't react to duplicate edge labels nor cycles
--   but throws errors on inserting edges to/from nodes not element of gr.
einsertEdge :: Earl -> (Node,Node,EdgeLabel) -> Earl
einsertEdge gr le@(a,s,_) | not (Graph.gelem a gr)    = error "ExtendedEarl.einsertEdge: Ancestor is not in graph."
                          | not (Graph.gelem s gr) = error "ExtendedEarl.einsertEdge: Successor is not in graph."
                          | otherwise              = Graph.insEdge le gr

-- | edeleteEdge gr (n,n'): deletes all edges (n,n') from the Earl gr regardless of the EdgeLabel
edeleteEdge :: Earl -> (Node,Node) -> Earl
edeleteEdge gr (n,n') = Graph.delEdge (n,n') gr

-- | edeleteLEdge gr le: deletes labeled edge form the Earl gr
edeleteLEdge :: Earl -> (Node,Node,EdgeLabel) -> Earl
edeleteLEdge gr le = Graph.delLEdge le gr

-- | buildEarl lns les: builds an Earl from the given labeled nodes lns and the given labeled edges les. If they
--   do not correspond, i.e. inserting a nowhere-leading edge, causes an error
buildEarl :: [(Node,Label)] -> [(Node,Node,EdgeLabel)] -> Earl
buildEarl = Graph.mkGraph