module Data.GraphRewriting.Grewrite where

import Data.GraphRewriting.Morphism
import Data.GraphRewriting.TermGraph
import Data.GraphRewriting.GraphRewriteRule
import Data.GraphRewriting.ExtendedEarl hiding (empty)
import Data.List as List (zip, find)
import Data.Maybe as Maybe (fromMaybe, fromJust)


-------------------------------------------------------------------------------------------------------------------
-- * outermost stuff
-------------------------------------------------------------------------------------------------------------------
      
-- | outermostrewrite grwsys tg: performs an outermost rewrite step on the given TermGraph tg with respect
--   to the given GraphRewriteSytem grwsys. Nothing if TermGraph tg is a Normal Form, i.e. no step is
--   is possible.
outermostrewrite :: GraphRewriteSystem -> TermGraph -> Maybe TermGraph
outermostrewrite grwsys tg = case findOutermostRedexPos grwsys maxsharedtg of
  Just (pos,rule) -> step rule sharedtg (nodeATpos sharedtg pos)
                     where sharedtg = unfold maxsharedtg pos
  Nothing         -> Nothing   
  where maxsharedtg = foldfromroot tg


-- | informativeoutermostrewrite grwsys tg: is a very informative outermost rewrite step which returns
--   (the rule used to rewrite, the position the step was taken at, the properly shared termgraph, the resulting termgraph)  
informativeoutermostrewrite :: GraphRewriteSystem -> TermGraph -> Maybe (GraphRewriteRule, Position, TermGraph, TermGraph)
informativeoutermostrewrite grwsys tg = case findOutermostRedexPos grwsys maxsharedtg of
  Just (pos, rule) -> Just (rule, pos, sharedTG, fromJust $ step rule sharedTG n)
                      where sharedTG = unfold maxsharedtg pos
                            n        = (nodeATpos sharedTG pos)
  Nothing          -> Nothing
  where maxsharedtg = foldfromroot tg



-- | findOutermostRedexPos grwsys tg: returns the position of the redex with the corresponing graphrewrite rule.
--   The redex is determined in TermGraph tg w.r.t. the GraphRewriteSystem grw.
findOutermostRedexPos :: GraphRewriteSystem -> TermGraph -> Maybe (Position, GraphRewriteRule)
findOutermostRedexPos grwsys tg = findRedex' grwsys (root tg) []
  where ftg = tg  -- maximally shared tg
        findRedex' (rule:rules) n pos | isRedex rule ftg n = Just (pos, rule)		                                  
                                      | otherwise          = findRedex' rules n pos  -- check all rules for node n		
        findRedex' []           n pos = findRedexInSuccs (succs ftg n) 1  -- check succs of n, 1 indicates fst successor
          where findRedexInSuccs []      _ = Nothing      
                findRedexInSuccs (si:ss) i = case findRedex' grwsys si (pos++[i]) of 
                                                   Just (pos',rule') -> Just (pos',rule') 
                                                   Nothing           -> findRedexInSuccs ss (i+1)

-------------------------------------------------------------------------------------------------------------------
-- * a rewrite step
-------------------------------------------------------------------------------------------------------------------
      
-- | step rule S n: performs a single rewrite step at S|^n with given rule. 
--   Nothing, if no morphism between lefthandside of rule and S|^n is found.
step :: GraphRewriteRule -> TermGraph -> Node -> Maybe TermGraph
step rule s n = case match rule' s n of
  Nothing       -> Nothing
  Just morphism -> Just (apply morphism (s,n) (rhs rule')) 
  where rule' = rename rule s

-- | match S T n: constructs mapping (i.e. a morphism) from node numbers in S|^n to node numbers in T.                             
match :: GraphRewriteRule -> TermGraph -> Node -> Maybe Morphism
match l s n = matchAt empty (l, rootlhs l) (s, n)

-- | matchAt morphism (S,n) (T,n'): extends morphism by (n,n'), if legal, and continues with matching succsessor respectively 
matchAt :: Morphism -> (GraphRewriteRule, Node) -> (TermGraph, Node) -> Maybe Morphism
matchAt morphism (pt, p) (s, n) = case context pt p of 
  (_, _, GVar _    , _)    -> extend morphism (p,n)
  (_, _, f@(GFun _), psuc) | f /= g    -> Nothing 
                           | otherwise -> case sortsuccs psuc nsuc of
                                            Nothing     -> Nothing
                                            Just (psuc',nsuc') -> matchAt' (extend morphism (p,n)) psuc' nsuc' 
    where (_, _, g, nsuc) = context s n
          matchAt' Nothing _ _   = Nothing
          matchAt' m [] []       = m
          matchAt' (Just morph) (pi:ps) (ni:ns) = matchAt' (matchAt morph (pt,pi) (s,ni)) ps ns
          matchAt' _ [] (_:_) = Nothing
          matchAt' _ (_:_) [] = Nothing
          
-- sorts the succsessor lists i.e. sorts nsuc according to psuc's edge labels 
-- i.e. sortsuccs [(1,p1),(2,p2)] [(2,n2),(1,n1)] = ([p1,p2],[n1,n2])
-- this is to enhance the the match function - to make sure they have the same successors (espacially amount of successors)
sortsuccs :: [(EdgeLabel, Node)] -> [(EdgeLabel, Node)] -> Maybe ([Node], [Node])
sortsuccs psuc nsuc = sort' psuc ([], []) nsuc 
  where sort' [] sorted [] = Just sorted
        sort' _  _      [] = Nothing
        sort' [] _      _  = Nothing
        sort' ((elp,p):ps) (pps, sortedns) (_:helpi') = case List.find (\(el,_) -> el == elp) nsuc of
          Just (_, n) -> sort' ps ((p:pps), (n:sortedns)) helpi'
          Nothing     -> Nothing




-- | apply morph (S,n) R: S|^n is replaced by m(R), where m(R) denotes to the application of the morphism morph on R.  
--   The application of the morphism maps a node in R either to the same node in R or to a node in S (when the node in R represented a variable node)    
apply :: Morphism -> (TermGraph, Node) ->  TermGraph ->  TermGraph
apply morphism (s,n) r | n ==  root s  = s' |^ m (root r)
                       | otherwise     = redirect s' n (m (root r)) |^ root s
  where s'     = foldd f r (root r) (s `union` r) 
        f n' si = redirect si n' (m n')
        m      = mstar morphism                             
        union u t = insertEdges (insertNodes u (lnodes t)) (ledges t) 


-- | rename rule TG: renames RULE wrt. to TG, i.e. changes node numbering in rule, such that (node numbers rule) ∩ (node numbers TG) = ø
rename :: GraphRewriteRule -> TermGraph -> GraphRewriteRule
rename rule tg = GR (m (rootlhs rule)) (m (rootrhs rule)) (buildEarl nodes' edges')
  where morphism = Maybe.fromMaybe (error "Grewrite.rename - sthg. wrong with renamning morphism") (fromList $ List.zip (nodes rule) nodesNo)
        nodesNo  = freshNodes tg (size rule)
        nodes'   = [(m n, lab) | (n,lab) <- lnodes rule]
        edges'   = [(m a, m s, l) | (a,s,l) <- ledges rule]
        m        = mstar morphism


-- | isRedex grw tg n: True if Node in Termgraph tg represents a redex w.r.t. GraphRewriteSystem grw.
--   Attention: The result False does not imply that n is not a redex as there might
--   be some sharing issues, with the precondition that tg is maximally shared, this implication holds however.
isRedex :: GraphRewriteRule -> TermGraph -> Node -> Bool
isRedex grw tg n = case match grw tg n of
  Just _  -> True
  Nothing -> False

{- old version, more readable - but less efficient
matchAt morphism (pt,p) (s,n) | isVarNode pt p               = extend morphism (p,n) 
                              | not (sameLabel (pt,p) (s,n)) = Nothing
                              | otherwise                    = matchAt' (extend morphism (p,n)) (succs pt p) (succs s n)
                                where matchAt' Nothing _ _   = Nothing
                                      matchAt' _ (_:_) []  = Nothing
                                      matchAt' _ [] (_:_)  = Nothing
                                      matchAt' m [] []       = m
                                      matchAt' (Just morph) (pi:ps) (ni:ns) = matchAt' (matchAt morph (pt,pi) (s,ni)) ps ns
-}