{-# LANGUAGE DeriveDataTypeable #-} 

module Main (
  TermRewriteSystem
  , grsfromfile
  , tgfromstring
  , grewrite
  , main ) where


import Data.Rewriting.Rule (Term, Rule)
import Data.Rewriting.Problem (parseFileIO, rules, strictRules)
import Data.Rewriting.Term (fromString, prettyTerm)
import qualified Data.Rewriting.Rules as TermRewriting (outerRewrite, result)

import Data.GraphRewriting.GraphRewriteRule
import Data.GraphRewriting.Grewrite
import Data.GraphRewriting.TermGraph
import Data.GraphRewriting.Printtt


import Text.PrettyPrint.ANSI.Leijen as Doc
import System.Console.CmdArgs.Implicit

-- | represents a term rewrite system 
type TermRewriteSystem = [Rule String String]

-- | record for command line arguments
data GraphRewrite = GraphRewriting {   termrewritesystem  :: String
                                     , term :: String
                                     , nf   :: Bool
                                     , trw  :: Bool
                                     } deriving (Show,Typeable,Data)

-- | grsfromfile fp: parses the term rewrite system from given file path fp and
-- returns the corresponding graphrewrite system and term rewrite system as a
-- pair
grsfromfile :: FilePath -> IO (GraphRewriteSystem, TermRewriteSystem)
grsfromfile fp = 
   parseFileIO fp >>= prob2sys 
   where prob2sys prob = return (toGraphRewriteSystem trs, trs)
           where trs = (strictRules (rules prob))  

-- | tgfromstring s: parses a minimally sharing termgraph from the given String s (which
-- corresponds to a term)
tgfromstring :: String -> IO (TermGraph, (Term String String))
tgfromstring st  = return (
  case (Data.Rewriting.Term.fromString (findVars st) st) of
    Left err -> error $ show err
    Right t  -> (toMinTermGraph t,t)
  )

-- finds all variables in the given (term) String (indicated by not having paranthesis)
findVars :: String -> [String]
findVars input = findV' input [] []
  where findV'  []    pot fvars = ((strip pot) `mycon` fvars)
        findV' (s:ss) pot fvars | s == '('              = findV' ss [] fvars
                                | s == ',' || s == ')'  = findV' ss [] ((strip pot) `mycon` fvars)
                                | otherwise             = findV' ss (pot ++ [s]) fvars
        mycon l ll = if (l /= "") then l:ll else ll 
        strip l = [ x | x <- l, x /= ' ']


-- | command line arguments 
grewrite :: Mode (CmdArgs GraphRewrite)
grewrite = cmdArgsMode $ GraphRewriting { 
    termrewritesystem = def &= argPos 0 &=typ "TRS"
  , term = def &= argPos 1 &= typ "TERM"
  , nf = def   &= help "only print normal form"   
--  , on = def   &= help "do not print any output" 
  , trw  = def &= help "use term-rewriting instead of graph-rewriting" 
  }
  &= summary "Programm to simulate Term Graph Rewriting."

main :: IO ()
main = do 
       arguments    <-  cmdArgsRun grewrite
       (grwsys,trs) <-  grsfromfile $ termrewritesystem arguments
       (tg,t)       <-  tgfromstring $ term arguments 
       result <- if (nf arguments)      
                 then if (trw arguments)  
                         then nftrw trs t                 -- nf & termrewriting
                         else nfgrw grwsys tg             -- nf & graphrewriting 
                 else if (trw arguments)
                         then stepstrw trs t              -- steps & termrewriting
                         else stepsgrw grwsys tg          -- graphrewriting & steps
       return result

-- The message to be printed if a term is in normal form
inNFmsg :: IO ()
inNFmsg = do  putStrLn "Term is already in Normal Form. \n"
              putStrLn "If you think, that this is not true, check"
              putStrLn "    a) whether you loaded the rewriting system correctly."
              putStrLn "    b) if you gave the term correctly, be aware that the programm doesn't do any arity checks for function symbols."
              putStrLn "    c) if a) and b) don't hold you probably found a problem, please contact csag9385@uibk.ac.at"
 
---------------------------------------------------------          
-- Term Rewriting
---------------------------------------------------------          

-- normal form term rewriting 
nftrw :: TermRewriteSystem -> Term String String -> IO ()
nftrw trs t = case TermRewriting.outerRewrite trs t of
  []   -> inNFmsg
  tn:_ -> putDoc (prettyTerm text text t) >> nftrw' (TermRewriting.result tn) 1  -- i => count steps, init value 1
  where nftrw' ti i = case TermRewriting.outerRewrite trs ti of
          []     -> putDoc $ text "->" <> integer i <> line <> (prettyTerm text text ti) <> line 
          tj:_  -> nftrw' (TermRewriting.result tj) (i+1)

-- stepwise normal form term rewriting
stepstrw :: TermRewriteSystem -> Term String String -> IO ()
stepstrw trs t =  case TermRewriting.outerRewrite trs t of
  []   -> inNFmsg
  tn:_ ->  putDoc (prettyTerm text text t) >> printi (TermRewriting.result tn) 1 >> stepstrw' (TermRewriting.result tn) 2 -- i => count steps, init value 1
  where printi tpr c = putDoc $ line <> text "->" <> integer c <> line <> (prettyTerm text text tpr)
        stepstrw' ti i = case TermRewriting.outerRewrite trs ti of
          []     ->  putDoc line
          tj:_  -> printi (TermRewriting.result tj) i >> stepstrw' (TermRewriting.result tj) (i+1)



---------------------------------------------------------          
-- Term Graph Rewriting
---------------------------------------------------------          

-- normal form graph rewriting
nfgrw :: GraphRewriteSystem -> TermGraph -> IO ()
nfgrw grwsys tg = case (outermostrewrite grwsys tg) of 
  Nothing   -> inNFmsg
  Just tg'  -> {-putDoc (printtg tg) >> -} nfgrw' tg' 1 
  where nfgrw' tg' n = case outermostrewrite grwsys tg' of
          Nothing   -> putDoc $ (text "=>") <> (integer n) <> line <> printtg tg' <>  line
          Just tg'' -> nfgrw' tg'' (n+1)

-- stepwise normal form graph rewriting
stepsgrw :: GraphRewriteSystem -> TermGraph -> IO ()
stepsgrw grwsys tg = case (outermostrewrite grwsys tg) of 
  Nothing   -> inNFmsg
  Just tg'  -> putDoc (printtg tg) >> printi tg' 1 >> (stepsgrw') tg' 2  -- i => count steps, init value 1  
  where printi  tgp c = (putDoc (line <> text "=>" <> integer c <> line <> printtg tgp))
        stepsgrw' tgi i = case outermostrewrite grwsys tgi of
          Nothing   -> putDoc line
          Just tgj -> printi tgj i >> (stepsgrw') tgj (i+1)  